# Generated by Django 2.2.6 on 2019-11-08 05:52

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_randomization", "0003_auto_20191024_0426")]

    operations = [
        migrations.AlterField(
            model_name="randomizationlist",
            name="allocated_site",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="sites.Site",
            ),
        )
    ]
