# Generated by Django 2.2.9 on 2020-02-02 20:01

import django_crypto_fields.fields.encrypted_char_field
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_randomization", "0004_auto_20191108_0752"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalrandomizationlist",
            name="assignment",
            field=django_crypto_fields.fields.encrypted_char_field.EncryptedCharField(
                blank=True, help_text=" (Encryption: RSA local)", max_length=71
            ),
        ),
        migrations.AlterField(
            model_name="randomizationlist",
            name="assignment",
            field=django_crypto_fields.fields.encrypted_char_field.EncryptedCharField(
                blank=True, help_text=" (Encryption: RSA local)", max_length=71
            ),
        ),
    ]
